@@ -1,16 +1,19 @@
+# encoding: utf8
+#
 # Physical Quantities
 #
 # Utilities for working with physical quantities (numbers with units).
 #
 # This module inputs and outputs in several different forms:
 #    quantities (numbers with units):
-#        (1e6, 'Hz') -- quant
-#        '1MHz'      -- eng_quant
-#        '1e6Hz'     -- flt_quant
+#        (1e6, 'Hz') -- quantity
+#        '1MHz'      -- engineering format
+#        '1e6Hz'     -- string
 #    numbers (numbers without units):
 #        1e6         -- num
-#        '1M'        -- eng_num
-#        '1e6'       -- flt_num
+#        '1M'        -- unitless engineering format
+#        '1e6'       -- unitless string
+#
 
 # License {{{1
 # Copyright (C) 2016 Kenneth S. Kundert
@@ -29,7 +32,7 @@
 # along with this program.  If not, see [http://www.gnu.org/licenses/].
 
 # Imports {{{1
-__version__ = '1.0.0'
+__version__ = '1.2.0'
 import re
 
 # Parameters {{{1
@@ -60,33 +63,33 @@ CONSTANTS = {
 
 # Constants {{{1
 MAPPINGS = {
-    'Y' : ('e24',  1e24 ),
-    'Z' : ('e21',  1e21 ),
-    'E' : ('e18',  1e18 ),
-    'P' : ('e15',  1e15 ),
-    'T' : ('e12',  1e12 ),
-    'G' : ('e9',   1e9  ),
-    'M' : ('e6',   1e6  ),
-    'K' : ('e3',   1e3  ),
-    'k' : ('e3',   1e3  ),
-    '_' : ('',     1    ),
-    ''  : ('',     1    ),
-    'm' : ('e-3',  1e-3 ),
-    'u' : ('e-6',  1e-6 ),
-    'n' : ('e-9',  1e-9 ),
-    'p' : ('e-12', 1e-12),
-    'f' : ('e-15', 1e-15),
-    'a' : ('e-18', 1e-18),
-    'z' : ('e-21', 1e-21),
-    'y' : ('e-24', 1e-24),
+    'Y': ('e24',  1e24 ),
+    'Z': ('e21',  1e21 ),
+    'E': ('e18',  1e18 ),
+    'P': ('e15',  1e15 ),
+    'T': ('e12',  1e12 ),
+    'G': ('e9',   1e9  ),
+    'M': ('e6',   1e6  ),
+    'K': ('e3',   1e3  ),
+    'k': ('e3',   1e3  ),
+    '_': ('',     1    ),
+    '' : ('',     1    ),
+    'm': ('e-3',  1e-3 ),
+    'u': ('e-6',  1e-6 ),
+    'n': ('e-9',  1e-9 ),
+    'p': ('e-12', 1e-12),
+    'f': ('e-15', 1e-15),
+    'a': ('e-18', 1e-18),
+    'z': ('e-21', 1e-21),
+    'y': ('e-24', 1e-24),
 }
 
 BIG_SCALE_FACTORS = 'kMGTPEZY'
-    # these must be given in order, one for every three decades
-    # use k rather than K, because K looks like a temperature when used alone
-    #
+    # These must be given in order, one for every three decades.
+    # Use k rather than K, because K looks like a temperature when used alone.
+
 SMALL_SCALE_FACTORS = 'munpfazy'
-    # these must be given in order, one for every three decades
+    # These must be given in order, one for every three decades.
 
 # Pattern Definitions {{{1
 # Build regular expressions used to recognize quantities
@@ -102,7 +105,7 @@ units = named_regex('units', r'(?:[a-zA-Z][-^/()\w]*)?')
 smpl_units = named_regex('units', r'(?:[a-zA-Z_]*)')
     # may only contain alphabetic characters, ex: V, A, Ohms, etc.
 currency = named_regex('currency', '[%s]' % CURRENCY_SYMBOLS)
-nan = named_regex('nan', '(?i)inf|nan')
+nan = named_regex('nan', '(?i:inf|nan)')
 left_delimit = r'(?:\A|(?<=[^a-zA-Z0-9_.]))'
 right_delimit = r'(?=[^-+0-9_]|\Z)'
 
@@ -183,8 +186,8 @@ simple_nan = (
 )
 
 all_number_converters = [
-    (re.compile('\A\s*{}\s*\Z'.format(pattern)), get_mant, get_sf, get_units)
-    for pattern, get_mant, get_sf, get_units  in [
+    (re.compile(r'\A\s*{}\s*\Z'.format(pattern)), get_mant, get_sf, get_units)
+    for pattern, get_mant, get_sf, get_units in [
         number_with_exponent, number_with_scale_factor, simple_number,
         currency_with_exponent, currency_with_scale_factor, simple_currency,
         nan_with_units, currency_nan, simple_nan,
@@ -192,14 +195,15 @@ all_number_converters = [
 ]
 
 sf_free_number_converters = [
-    (re.compile('\A\s*{}\s*\Z'.format(pattern)), get_mant, get_sf, get_units)
-    for pattern, get_mant, get_sf, get_units  in [
+    (re.compile(r'\A\s*{}\s*\Z'.format(pattern)), get_mant, get_sf, get_units)
+    for pattern, get_mant, get_sf, get_units in [
         number_with_exponent, simple_number,
         currency_with_exponent, simple_currency,
         nan_with_units, currency_nan, simple_nan,
     ]
 ]
 
+# Regular expression for recognizing and decomposing string .format method codes
 format_spec = re.compile(r'\A([<>]?)(\d*)(?:\.(\d+))?([qruseEfFgGdnQR]?)\Z')
 
 # Utilities {{{1
@@ -244,7 +248,7 @@ AssignmentFormatter = DEFAULT_ASSIGNMENT_FORMATTER
 AssignmentRecognizer = re.compile(DEFAULT_ASSIGNMENT_RECOGNIZER)
 
 def set_preferences(
-        hprec=False, mprec =False, spacer=False, unity=False, output=False,
+        hprec=False, mprec=False, spacer=False, unity=False, output=False,
         ignore_sf=0, assign_fmt=False, assign_rec=False
 ):
     """Set Global Preferences
@@ -263,8 +267,9 @@ def set_preferences(
         familiar scale factors.
     ignore_sf (bool): Whether scale factors should be ignored by default.
     assign_fmt (str): Format string for an assignment. Will be passed through
-        .format function. Format string takes three possible arguments named n,
-        q, and d for the name, value and description.  The default is '{n} = {v}'
+        string .format method. Format string takes three possible arguments
+        named n, q, and d for the name, value and description.  The default is
+        '{n} = {v}'
     assign_rec (str): Regular expression used to recognize an assignment. Used
         in add_to_namespace(). Default recognizes the form
             "Temp = 300_K -- Temperature".
@@ -291,7 +296,7 @@ def set_preferences(
         OutputScaleFactors = (
             output if output is not None else DEFAULT_OUTPUT_SCALE_FACTORS
         )
-    if ignore_sf is not 0:
+    if ignore_sf != 0:
         IgnoreScaleFactors = (
             ignore_sf if ignore_sf is not None else DEFAULT_IGNORE_SCALE_FACTORS
         )
@@ -305,8 +310,8 @@ def set_preferences(
         )
 
 # Quantity class {{{1
-class Quantity:
-    def __init__(self, value, units=None, ignore_sf=None):
+class Quantity(float):
+    def __new__(cls, value, units='', ignore_sf=None):
         """Physical Quantity
         A real quantity with units.
 
@@ -315,10 +320,6 @@ class Quantity:
                 2.5ns, 1.7 MHz, 1e6ohms, 2.8_V, 1e12 F, 42, etc.
         units: the quantities units.
         """
-        if units is None:
-            units = ''
-        self._value = value
-        self.units = units
         ignore_sf = IgnoreScaleFactors if ignore_sf is None else ignore_sf
 
         if is_str(value):
@@ -326,41 +327,57 @@ class Quantity:
                 number_converters = sf_free_number_converters
             else:
                 number_converters = all_number_converters
-            # if we get a string, keep all the pieces so we can reconstruct it
-            # exactly as it was given.
             for pattern, get_mant, get_sf, get_units in number_converters:
                 match = pattern.match(value)
                 if match:
-                    self._value = None
-                    self._mantissa = get_mant(match)
+                    mantissa = get_mant(match)
                     sf = get_sf(match)
-                    self._scale_factor = sf if sf != '_' else ''
-                    self.units = get_units(match)
-                    if self.units:
-                        if units:
-                            assert units == self.units
+                    sf = sf if sf != '_' else ''
+                    if units:
+                        assert units == get_units(match), 'mismatched units'
                     else:
-                        self.units = units
-                    return
-            try:
-                self._value, self.units = CONSTANTS[value]
-            except KeyError:
-                raise ValueError('%s: not a valid number.' % value)
+                        units = get_units(match)
+                    number = float(mantissa + MAPPINGS.get(sf, [sf])[0])
+                    break
+            else:
+                try:
+                    number, units = CONSTANTS[value]
+                except KeyError:
+                    raise ValueError('%s: not a valid number.' % value)
+        else:
+            number = value
+
+        self = float.__new__(cls, number)
+        self.units = units
+        try:
+            # if we got a string, keep the pieces so we can reconstruct it
+            # exactly as it was given.
+            self._mantissa = mantissa
+            self._scale_factor = sf
+        except NameError:
+            pass
+        return self
 
     def is_infinite(self):
-        value = self._mantissa if self._value is None else str(self._value)
+        try:
+            value = self._mantissa
+        except AttributeError:
+            value = str(self.real)
         return value.lower() in ['inf', '-inf', '+inf']
 
     def is_nan(self):
-        value = self._mantissa if self._value is None else str(self._value)
+        try:
+            value = self._mantissa
+        except AttributeError:
+            value = str(self.real)
         return value.lower() in ['nan', '-nan', '+nan']
 
     def add_name(self, name):
-        """Add a name."""
+        "Add a name."
         self.name = name
 
     def add_desc(self, desc):
-        """Add a description."""
+        "Add a description."
         self.desc = desc
 
     def strip(self):
@@ -370,18 +387,18 @@ class Quantity:
         access to the value specified without any loss of precision if the value
         was specified as a string.
         """
-        if self._value is None:
+        try:
             return self._mantissa + self._scale_factor
-        else:
-            return num_to_str(self._value)
+        except AttributeError:
+            return num_to_str(self.real)
 
     def to_float(self):
         """Returns the value as a float."""
-        if self._value is None:
+        try:
             sf = self._scale_factor
             return float(self._mantissa + MAPPINGS.get(sf, [sf])[0])
-        else:
-            return self._value
+        except AttributeError:
+            return self.real
 
     def to_unitless_str(self):
         """Renders the value as a string in floating point notation.
@@ -391,14 +408,14 @@ class Quantity:
         specified without any loss of precision if the value was specified as a
         string.
         """
-        if self._value is None:
+        try:
             sf = self._scale_factor
             return self._mantissa + MAPPINGS.get(sf, [sf])[0]
-        else:
-            return num_to_str(self._value)
+        except AttributeError:
+            return num_to_str(self.real)
 
     def to_unitless_eng(self, prec=None):
-        """Renders the value as a string in engineering notation."""
+        "Renders the value as a string in engineering notation."
         # this is a bit of a hack, temporarily remove the units
         units = self.units
         self.units = None
@@ -407,17 +424,17 @@ class Quantity:
         return eng_number
 
     def to_tuple(self):
-        """Returns a tuple that contains the value as a float and the units."""
+        "Returns a tuple that contains the value as a float and the units."
         return self.to_float(), self.units
 
     def to_str(self):
-        """Renders the value and units as a string in floating point notation."""
+        "Renders the value and units as a string in floating point notation."
         number = self.to_unitless_str()
         units = self.units
         return _combine(number, '', units, Spacer)
 
     def to_eng(self, prec=None):
-        """Renders the value and units as a string in engineering notation."""
+        "Renders the value and units as a string in engineering notation."
 
         # determine precision
         if prec is None:
@@ -465,18 +482,48 @@ class Quantity:
             num = 100*float(mantissa)
         mantissa = "%.*f" % (prec-shift, num)
 
-        #remove trailing zeros (except if mantissa does not contain a .)
+        # remove trailing zeros (except if mantissa does not contain a .)
         if mantissa.find('.') >= 0:
             mantissa = mantissa.rstrip("0")
 
-        #remove trailing decimal point
+        # remove trailing decimal point
         mantissa = mantissa.rstrip(".")
 
         return _combine(mantissa, sf, units, Spacer)
 
+    def to_sci(self, prec=None):
+        "Renders the value and units as a string in scientific notation."
+
+        # determine precision
+        if prec is None:
+            prec = HumanPrecision
+        else:
+            prec = int(prec)
+        assert (prec >= 0)
+
+        # check for infinities or NaN
+        if self.is_infinite() or self.is_nan():
+            return _combine(self.strip(), '', self.units, ' ')
+
+        # convert into scientific notation with proper precision
+        value = self.to_float()
+        units = self.units
+        number = "%.*e" % (prec, value)
+        mantissa, exp = number.split("e")
+        exp = exp.replace('+', '')
+        superscripts = str.maketrans('-0123456789', '⁻⁰¹²³⁴⁵⁶⁷⁸⁹')
+        sf = '×10' + exp.translate(superscripts)
+        return _combine(mantissa, sf, units, Spacer)
+
+    def __float__(self):
+        return self.to_float()
+
     def __str__(self):
         return self.to_eng()
 
+    def __repr__(self):
+        return 'Quantity({!r})'.format(self.to_eng())
+
     def __format__(self, fmt):
         """Convert quantity to string for Python string format function.
 
@@ -488,7 +535,7 @@ class Quantity:
         A is character and gives the alignment: either '', '>', or '<'
         W is integer and gives the width
         P is integer and gives the precision
-        T is  and gives the type: choose from q, r, s, e, f, F, g, G, u, n, ...
+        T is char and gives the type: choose from q, r, s, e, f, g, u, n, d, ...
            q = quantity (1.4204GHz)
            r = real (1.4204G)
            s = string (1.4204GHz)
@@ -539,9 +586,6 @@ class Quantity:
         else:
             return self.to_eng()
 
-    def __float__(self):
-        return self.to_float()
-
 
 # Shortcut functions {{{1
 def quant_to_tuple(value, units=None):
@@ -550,6 +594,9 @@ def quant_to_tuple(value, units=None):
 def quant_to_eng(value, units=None, prec=None):
     return Quantity(value, units).to_eng(prec)
 
+def quant_to_sci(value, units=None, prec=None):
+    return Quantity(value, units).to_sci(prec)
+
 def quant_to_str(value, units=None):
     return Quantity(value, units).to_str()
 
@@ -572,7 +619,6 @@ def all_to_eng_fmt(text):
 
     It is assumed that any units are assumed to be simple, meaning that they
     contain only alphabetic characters (no numbers or symbols)."""
-
     out = []
     start = 0
     for match in embedded_floating_point_notation.finditer(text):
@@ -580,7 +626,7 @@ def all_to_eng_fmt(text):
         number = match.group(0)
         try:
             number = quant_to_eng(number)
-        except ValueError: # pragma: no cover
+        except ValueError:  # pragma: no cover
             # something unexpected happened
             # but this is not essential, so ignore it
             pass
@@ -590,12 +636,11 @@ def all_to_eng_fmt(text):
 
 # All from engineering format {{{2
 def all_from_eng_fmt(text):
-    """Replace all occurrences of quantities found in text to engineering format.
+    """Convert all occurrences of quantities found in text to engineering format
 
     It is assumed that there is no space between the number and the scale factor
     and any units are assumed to be simple, meaning that they contain only
     alphabetic characters (no numbers or symbols)."""
-
     out = []
     start = 0
     for match in embedded_engineering_notation.finditer(text):
@@ -603,7 +648,7 @@ def all_from_eng_fmt(text):
         number = match.group(0)
         try:
             number = quant_to_str(number)
-        except ValueError: # pragma: no cover
+        except ValueError:  # pragma: no cover
             # something unexpected happened
             # but this is not essential, so ignore it
             pass
@@ -620,11 +665,11 @@ def add_to_namespace(quantities):
     """ Add to Namespace
 
     Takes a string that contains quantity definitions and places those
-    quantities in the calling namespace. The string may contain one definition per line, of the form:
+    quantities in the calling namespace. The string may contain one definition
+    per line, of the form:
         <name> = <value> -- <description>
     The description is discarded.
     """
-
     # Access the namespace of the calling frame
     import inspect
     frame = inspect.stack()[1][0]
@@ -642,5 +687,5 @@ def add_to_namespace(quantities):
             quantity.add_name(name)
             quantity.add_desc(desc)
             namespace[name] = quantity
-        else: # pragma: no cover
+        else:  # pragma: no cover
             raise ValueError('{}: not a valid number.'.format(line))
